#! /bin/gawk -f

# Script Arguments:
# - header  :: HTML header tag to use for title, defaults to h1.
# - datePos :: Where to put the date. Can be set to prebody or postbody.
#              Defaults to prebody.
# - timestamp :: Set to short for short form file timestamps (YYYYMMDD)
#                instead of full form (YYYY-MM-DD-Day-HH-MM).
#                Preferable in links.
# - maxHeaderLevel :: The maximum header level to use. Any headers with higher
#                     levels will be translated to ordered lists.
#                     Defaults to 3.
# - wrapHeadingBodies :: If set to true, wraps heading bodies in divs with the
#                        "heading-body" class, and gives headings the "heading"
#                        class.

# This script uses file descriptor 3 to return the timestamp.
# Example calls:
# ./export.awk ../src/news-raw/foo.org 3>/dev/null
# ./export.awk ../src/news-raw/foo.org 3>&1 1>/dev/null

# Available directives:
# #+TITLE: <title>
# #+DATE: <date> :: Date expected in <YYYY-MM-DD Day HH:MM> format
# #+META: <information>
# #+begin_src <lang> ... #+end_src :: Export source code
# #+begin ... #+end :: Export block contents literally

# Supports the following org mode markup:
# - headings
# - links
# - unordered lists with - +
# - ordered lists with 1. 1) a. A.
# - definition lists (-/+ term :: definition)
# - bold text with *
# - italic text with /
# - underlined text with _
# - strikethrough text with +
# - inline code with ~
# - one-liners with :
# - heading property blocks (translated into HTML attributes)

BEGIN {
  if (!header) {
    header = "h1"
  }
  if (!maxHeaderLevel) {
    maxHeaderLevel = 3
  }
  OpenListCount = 0
  IGNORECASE = 1  # Neither HTML nor org markup is case-sensitive
}

/^#\+TITLE:/ {
  html_encode()
  printf "<%s%s>%s</%s>\n", header, wrapHeadingBodies ? " class=\"heading\"" : "", substr($0, length("#+TITLE: ") + 1), header;
  if (wrapHeadingBodies) {
    print "<div class=\"heading-body\">"
    PrintedTitle = 1;
  }
  next;
}

/^#\+DATE:/ {
  dateFull = substr($0, length("#+DATE: <") + 1, length($0) - (length("#+DATE: <") + 1))
  dateShort = substr(dateFull, 1, length(dateFull) - 6)
  if (datePos != "postbody") {
    printf "<time datetime=\"%s\">%s</time>\n", substr(dateShort, 1, length(dateShort)-4), dateShort;
  }
  next;
}

/^#\+META:/ {
  metaData[metaDataCount++] = $0;
  next;
}

/^#\+begin/ {
  blockDepth++;
  if (blockDepth == 1) {
    if (withinHeading) {
      heading_print()
    }
    withinBlock = "literal";
    if ($1 == "#+begin_src") {
      withinBlock = "src"
      "mktemp" | getline srcFile
      close("mktemp")
      print "-*- mode: " $2 " -*-" > srcFile
    }
    next;
  }
}

/^#\+end/ {
  blockDepth--;
  if (blockDepth < 0) {
    print "Stray block ending" > "/dev/stderr"
    exit 1
  } else if (blockDepth == 0) {
    if (srcFile) {
      withinPre = 0
      fflush(srcFile)
      while (("./syntax-highlight.el " srcFile | getline) > 0) {
        if ($0 ~ "<pre>") {
          withinPre = 1;
        }
        if (withinPre) {
          print
        }
        if ($0 ~ "</pre>") {
          break;
        }
      }
      close ("./syntax-highlight.el" srcFile)
    }
    withinBlock = 0;
    srcFile = 0;
    next;
  }
}

{
  if (!withinBlock) {
    if ($0) {
      # Org to HTML translation
      if (withinHeading) {
        # Heading property block parsing
        heading_parse_properties()
      }
      lineType = determine_linetype()
      # Close any lists according to the indentation and line type
      list_close(lineType)
      if (linetype_is_list(lineType)) {
        # Headings and lists
        list_open(lineType)
        if (lineType ~ /[0-9]+/) {
          # Headings
          LatestHeading["level"] = lineType
          html_encode()
          LatestHeading["text"] = substr($0, index($0, $2))
          withinHeading = "HEAD"
        } else if (lineType == "dl") {
          # Definition lists
          export_text_line()
          print gensub("[-\\+] (.+) :: (.+)", "<dt>\\1</dt><dd>\\2</dd>", "g")
        } else {
          # Other lists
          export_text_line()
          print "<p>" substr($0, index($0, $2)) "</p>"
        }
      } else if (lineType == "p") {
        # Paragraphs
        export_text_line()
        print "<p>" $0 "</p>"
      } else {
        # Oneliners
        html_encode()
        print "<span class=\"inline-code\">" substr($0, index($0, $2)) "</span>"
      }
    }
  } else {
    if (srcFile) {
      # Code block
      print $0 > srcFile
    } else {
      # Generic block
      print $0
    }
  }
}

END {
  if (withinHeading) {
    heading_print()
  }
  list_close_all()
  if ((datePos == "postbody") && (dateShort)) {
    printf "<time datetime=\"%s\">%s</time>\n", substr(dateShort, 1, length(dateShort)-4), dateShort;
  }
  if (wrapHeadingBodies && PrintedTitle) {
    print "</div>"
  }
  # Print metadata
  if (metaDataCount) {
    print("<!--")
    for (i = 0; i < metaDataCount; i++) {
      print(metaData[i])
    }
    print("-->")
  }
  # Return timestamp via file descriptor 3 if a date is found
  if (dateFull) {
    # Sanitize dateFull so that it can be used as a file name later on.
    # Colons have to be replaced - they cause errors in make.
    gsub("[ :]", "-", dateFull);
    if (timestamp == "short") {
      dateFull = substr(dateFull, 1, 10)
      gsub("-", "", dateFull)
    }
    print dateFull | "cat 1>&3"
  }
}

function html_encode() {
  # Encode current line as a html-safe string
  gsub("&", "\\&amp;")
  gsub("<", "\\&lt;")
  gsub(">", "\\&gt;")
}

function export_text_line(_has_export_escape) {
  # Applies export escapes, encoding, links and text effects to current line
  _has_export_escape = index($0, "`");
  if (_has_export_escape) {
    apply_export_escape_mask();
  }
  html_encode();
  translate_links();
  apply_text_effects();
  if (_has_export_escape) {
    remove_export_escape_mask();
  }
}

function apply_export_escape_mask(_i) {
  # Surround with spaces to simplify regex
  $0 = " " $0 " "

  # Mark open-close-escaped delimiters
  delimiter_left  = "([[:space:]])`";
  delimiter_right = "`([[:space:][:punct:]])";
  delimiter_escaped = "\\\\`";
  gsub(delimiter_escaped, "");
  $0 = gensub(delimiter_left, "\\1", "g");
  $0 = gensub(delimiter_right, "\\1", "g");

  # Iterate over the string and mask delimited ranges
  _i = 1;
  while (match($0, "[^]*")) {
    ExportEscape[_i] = substr($0, RSTART+1, RLENGTH-2);
    $0 = replace_substr($0, RSTART, RLENGTH, "");
    _i++;
  }

  # Revert unused replacements
  gsub("", "`");
  gsub("", "`");
  gsub("", "`");

  # Remove the surrounding spaces
  $0 = substr($0, 2, length($0) - 2)
}

function remove_export_escape_mask(_i) {
  # Iterate over the string and unmask delimited ranges
  _i = 1;
  while (match($0, "")) {
    $0 = replace_substr($0, RSTART, RLENGTH, ExportEscape[_i]);
    _i++;
  }

  # Revert unused replacements
  gsub("", "`");
  gsub("", "`");
  gsub("", "`");
}

function translate_links(_link_contents_arr, _link_html, _link_html_format_str) {
  while (match($0, "\\[\\[([^\\]]*)\\]\\[([^\\]]*)\\]\\]", _link_contents_arr)) {
    if (_link_contents_arr[1] ~ "^id:.*") {
      # ID links (only used to link within the same file)
      _link_contents_arr[1] = substr(_link_contents_arr[1], 4)
      _link_html_format_str = "<a href=\"#%s\">%s</a>"
    } else if (_link_contents_arr[1] == sprintf("mailto:%s", _link_contents_arr[2])) {
      # Mail links with matching text content and href directives.
      # These are given the mail class.

      # CSS can trivially match links that are mailto: directives,
      # but not links whose text content is identical to the href
      # address they contain.
      # Marking such links allows them to be styled uniquely, e.g.
      # preventing line breaks at hyphens.
      _link_html_format_str = "<a href=\"%s\" class=\"mail\">%s</a>"
    } else {
      # Other links
      _link_html_format_str = "<a href=\"%s\">%s</a>"
    }
    _link_html = sprintf(_link_html_format_str, _link_contents_arr[1], _link_contents_arr[2])
    $0 = replace_substr($0, RSTART, RLENGTH, _link_html)
  }
}

function apply_text_effects() {
  # Surround with spaces to simplify regex
  $0 = " " $0 " "
  # Translate italic text
  $0 = apply_text_effect($0, "/", "/", "<em>", "</em>");
  # Translate bold text
  $0 = apply_text_effect($0, "\\*", "*", "<strong>", "</strong>");
  # Translate underlined text
  $0 = apply_text_effect($0, "_", "_", "<ins>", "</ins>");
  # Translate strikethrough text
  $0 = apply_text_effect($0, "\\+", "+", "<del>", "</del>");
  # Translate inline code
  $0 = apply_text_effect($0, "~", "~", "<span class=\"inline-code\">", "</span>");
  # Remove the surrounding spaces
  $0 = substr($0, 2, length($0) - 2)
}

function apply_text_effect(str, delimiter_regex, delimiter_raw,
                           wrap_begin, wrap_end) {
  # delimiter_regex: String representing delimiter as regex, e.g. \\* for *
  # delimiter_raw: String representing the raw delimiter character, e.g. *

  # Preprocess the string by replacing:
  # - left/right delimiters with ASCII shift-in, shift-out characters
  # - escaped delimiters with the ASCII data link escape character
  delimiter_left  = "([[:space:]])" delimiter_regex;
  delimiter_right = delimiter_regex "([[:space:][:punct:]])";
  delimiter_escaped = "\\\\" delimiter_regex;
  gsub(delimiter_escaped, "", str);
  str = gensub(delimiter_left, "\\1", "g", str);
  str = gensub(delimiter_right, "\\1", "g", str);

  # Apply the text effect
  str = gensub("([^]*)", wrap_begin "\\1" wrap_end, "g", str);

  # Revert unused replacements
  gsub("", delimiter_raw, str);
  gsub("", delimiter_raw, str);
  gsub("", delimiter_raw, str);

  return str;
}

#
# List Management Stuff
#

function list_open(type) {
  # Possible types: [0-9]+ (heading levels), ol, ul, dl
  # Can skip opening a list if it's detected to be a continuation of
  #   an existing list.

  # Calculate indentation for ol, ul, dl lists
  if (type ~ "^ol$|^ul$|^dl$") {
    indent = index($0, $1) - 1
  }

  if (OpenListCount &&
      (type == OpenLists[OpenListCount-1]["type"]) &&
      ((type !~ "^ol$|^ul$|^dl$") ||
       (indent == OpenLists[OpenListCount-1]["indent"]))) {
    # A continuation of the latest list
    if ((type ~ /[0-9]+/) && wrapHeadingBodies) {
      printf "</div>"
    }
    if ((type != "dl") && !((type ~ /[0-9]+/) && type <= maxHeaderLevel)) {
      print "</li><li>"
    }
    return
  }

  # Open the list
  if (type !~ /[0-9]+/ || type > maxHeaderLevel) {
    print "<" ((type ~ /[0-9]+/) ? "ol" : type) ">" ((type != "dl") ? "<li>" : "")
  }
  OpenLists[OpenListCount]["type"] = type
  OpenLists[OpenListCount]["indent"] = indent # Unused for headings
  OpenListCount++
}

function list_close(lineType) {
  # Closes all existing HTML list elements depending on indentation and line type.

  # Calculate current indentation
  indent = index($0, $1) - 1

  # Close any lists as required
  for (i = OpenListCount - 1; i >= 0; i--) {
    if ((lineType ~ /[0-9]+/)) {
      # Headings close all lists that aren't a heading with a lower level
      if ((OpenLists[i]["type"] ~ "^ol$|^ul$|^dl$") ||
          (OpenLists[i]["type"] > lineType)) {
        close_latest_list();
      } else {
        break;
      }
    } else if (lineType ~ "^ol$|^ul$|^dl$") {
      # ol, ul, dl lists close other ol, ul, dl lists with higher indentation
      if ((OpenLists[i]["type"] ~ "^ol$|^ul$|^dl$") &&
          ((OpenLists[i]["indent"] > indent) ||
           ((OpenLists[i]["indent"] == indent) &&
            (OpenLists[i]["type"] != lineType)))) {
        close_latest_list()
      } else {
        break;
      }
    } else {
      # Anything else closes all ol, ul, dl lists with higher or equal indentation
      if (((OpenLists[i]["type"] ~ "^ol$|^ul$|^dl$") &&
           (OpenLists[i]["indent"] >= indent))) {
        close_latest_list()
      } else {
        break;
      }
    }
  }
}

function list_close_all() {
  # Close all open lists unconditionally
  for (i = OpenListCount - 1; i >= 0; i--) {
    close_latest_list()
  }
}

function close_latest_list(_i, _type) {
  # Called for all types of lists, including low level headers.
  _i = OpenListCount - 1
  _type = OpenLists[_i]["type"]
  if (wrapHeadingBodies && _type ~ /[0-9]+/) {
    printf "</div>"
  }
  if (_type !~ /[0-9]+/ || _type > maxHeaderLevel) {
    print ((OpenLists[_i]["type"] != "dl") ? "</li>" : "") "</" ((OpenLists[_i]["type"] ~ /[0-9]+/) ? "ol" : OpenLists[_i]["type"]) ">"
  }
  delete OpenLists[_i]
  OpenListCount--
}

function determine_linetype(_type) {
  # If line is a list item, returns ul, ol, dl or [0-9]+ depending on its type.
  # If not, returns p for paragraph, oneliner for oneliners.
  if ($0 ~ /^\*+ /) {
    # Headings, calculate level
    _type = length($1) + substr(header, 2, 1)
  } else if ((($1 == "-") && ($0 ~ /- .+ :: .+/)) ||
             (($1 == "+") && ($0 ~ /\+ .+ :: .+/))) {
    # Definition lists
    _type = "dl"
  } else if (($1 == "-") || ($1 == "+")) {
    # Unordered lists
    _type = "ul"
  } else if (($1 ~ /^[0-9]+\.$/) || ($1 ~ /^[0-9]+)$/) ||
             ($1 ~ /^[a-z]{1}\.$/) || ($1 ~ /^[a-z]{1})$/)) {
    # Ordered lists
    _type = "ol"
  } else if ($1 == ":") {
    # Oneliners
    _type = "oneliner"
  } else {
    # Paragraphs
    _type = "p"
  }
  return _type
}

function linetype_is_list(_type) {
  # Returns true if given type represents an org list or heading
  return (_type ~ "^ol$|^ul$|^dl$") || (_type ~ /[0-9]+/);
}

#
# Heading Management Stuff
#

function heading_parse_properties() {
  # Called after a heading is detected to parse heading property blocks.

  if (withinHeading == "HEAD") {
    if ($0 == ":PROPERTIES:") {
      # The heading contains a properties block, prepare to parse it.
      withinHeading = "PROP"
      next;
    } else {
      # The heading does not contain a properties block. Just print it.
      heading_print()
    }
  } else if (withinHeading == "PROP") {
    if ($0 == ":END:") {
      # Heading property block ended, print
      heading_print()
      next;
    } else {
      # Parse property by transforming it e.g. from :ID: to id
      LatestHeading[substr(tolower($1), 2, length($1) - 2)] = substr($0, index($0, $2))
      next;
    }
  }
}

function heading_print(_headingTag) {
  # Prints the latest heading

  _headingTag = (LatestHeading["level"] > maxHeaderLevel) ? "p" : "h" LatestHeading["level"]

  printf "<" _headingTag
  if (wrapHeadingBodies || LatestHeading["class"]) {
    # Handle the class attribute of the heading first, as we may append to it.
    printf " class=\"" (wrapHeadingBodies ? "heading " : "") LatestHeading["class"] "\""
  }
  for (attr in LatestHeading) {
    if (attr != "text" && attr != "level" && attr != "class") {
      printf " %s=\"%s\"", attr, LatestHeading[attr]
    }
  }
  printf ">"
  printf LatestHeading["text"]
  print "</" _headingTag ">"
  if (wrapHeadingBodies) {
    print "<div class=\"heading-body\">"
  }
  delete LatestHeading
  withinHeading = 0
}

# Miscellaneous

function replace_substr(str, begin, len, new) {
  return substr(str, 1, begin-1) new substr(str, begin+len);
}
